import configparser
import os

class GITRepository(object):
    worktree = None
    conf = None
    gitdir = None

    def __init__(self, path, force=False):
        from helpers.repo.helpers import repo_file  # Lazy importing to avoid circular import conflicts
        self.worktree = path
        self.gitdir = os.path.join(path, '.git')

        if not (force or os.path.isdir(self.gitdir)):
            raise Exception(".git doesnt exist")

        # Reading configuration file in the .git repository
        self.conf = configparser.ConfigParser()
        cf = repo_file(self, "config")

        if cf:
            self.conf.read([cf])
        elif not force:
            raise Exception("No config path present")
        
        if not force:
            vers = int(self.conf.get("core", "repositoryformatversion"))
            if vers != 0:
                raise Exception("Version is unsupported")



from argparsing import argparser
import sys
from cmds.commands import *
from termcolor import cprint
# from stream import *


def main(argvs=sys.argv[1:]):
    args = argparser.parse_args(argvs)

    cprint(f"zyra rolling...", (196, 251, 174), "on_black", ["bold", "blink"])
    print("\033[38;5;81m")

    match args.command:
        case "add":
            cmd_add(args)
        case "checkout":
            cmd_checkout(args)
        case "init":
            cmd_init(args)
        case "cat-file":
            cmd_cat_file(args)
        case "hash-object":
            cmd_hash_obj(args)
        case "log":
            cmd_log(args)
        case "show-ref":
            cmd_show_ref(args)
        case "tag":
            cmd_tag(args)
        case "rev-parse":
            cmd_rev_parse(args)
        case "status":
            cmd_status(args)
        case "rm":
            cmd_rm(args)
        case "commit":
            cmd_commit(args)
        case "all-commits":
            cmd_commits(args)
        case "branch":
            cmd_branch(args)



from helpers.objects.helpers import object_find
from common.tag.tag_obj import GITTag
from common.objects import GITObject
from helpers.refs.helpers import ref_create
from common.objects import GITObject


def tag_create(repo, name, ref, create_tag_object=False):
    sha = object_find(repo, ref)
    obj = GITObject.object_read(repo, sha)

    if create_tag_object:
        tag = GITTag()
        tag.kvlm = dict()
        tag.kvlm[b"object"] = sha.encode()
        tag.kvlm[b"type"] = obj.obj_type 
        tag.kvlm[b"tag"] = name.encode()
        tag.kvlm[b"tagger"] = b"Someone <someone@example.com>"
        tag.kvlm[None] = (
            b"This tag is generated by the zyra\n"
        )
        tag_sha = GITObject.object_write(tag, repo)
        ref_create(repo, "tags/" + name, tag_sha)
    else:
        ref_create(repo, "tags/" + name, sha)



import os
from helpers.repo.helpers import repo_file
from stage.indexfile import GITIndex, GITIndexEntry
from math import ceil

def index_write(repo, index):
        with open(repo_file(repo, "index"), "wb") as f:
            f.write(b"DIRC")
            f.write(index.version.to_bytes(4, "big"))
            f.write(len(index.entries).to_bytes(4, "big"))

            idx = 0
            for e in index.entries:
                f.write(e.ctime[0].to_bytes(4, "big"))
                f.write(e.ctime[1].to_bytes(4, "big"))
                f.write(e.mtime[0].to_bytes(4, "big"))
                f.write(e.mtime[1].to_bytes(4, "big"))
                f.write(e.dev.to_bytes(4, "big"))
                f.write(e.ino.to_bytes(4, "big"))

                mode = (e.mode_type << 12) | e.mode_perms
                f.write(mode.to_bytes(4, "big"))
                f.write(e.uid.to_bytes(4, "big"))
                f.write(e.gid.to_bytes(4, "big"))

                f.write(e.fsize.to_bytes(4, "big"))
                f.write(int(e.sha, 16).to_bytes(20, "big"))
                flag_assume_valid = 0x1 << 15 if e.flag_assume_valid else 0
                name_bytes = e.name.encode("utf8")
                bytes_len = len(name_bytes)
                if bytes_len >= 0xFFF:
                    name_length = 0xFFF
                else:
                    name_length = bytes_len

                f.write((flag_assume_valid | e.flag_stage | name_length).to_bytes(2, "big"))

                f.write(name_bytes)
                f.write((0).to_bytes(1, "big"))

                idx += 62 + len(name_bytes) + 1
                if idx % 8 != 0:
                    pad = 8 - (idx % 8)
                    f.write((0).to_bytes(pad, "big"))
                    idx += pad

def index_read(repo):
    index_file = repo_file(repo, "index")

    if not os.path.exists(index_file):
        return GITIndex()

    with open(index_file, 'rb') as f:
        raw = f.read()

    header = raw[:12]
    signature = header[:4]
    assert signature == b"DIRC" # -> "DirCache"
    version = int.from_bytes(header[4:8], "big")
    assert version == 2, "wyag only supports index file version 2"
    count = int.from_bytes(header[8:12], "big")

    entries = list()

    content = raw[12:]
    idx = 0
    for i in range(0, count):
        ctime_s =  int.from_bytes(content[idx: idx+4], "big")
        ctime_ns = int.from_bytes(content[idx+4: idx+8], "big")
        mtime_s = int.from_bytes(content[idx+8: idx+12], "big")
        mtime_ns = int.from_bytes(content[idx+12: idx+16], "big")
        dev = int.from_bytes(content[idx+16: idx+20], "big")
        ino = int.from_bytes(content[idx+20: idx+24], "big")
        unused = int.from_bytes(content[idx+24: idx+26], "big")
        assert 0 == unused
        mode = int.from_bytes(content[idx+26: idx+28], "big")
        mode_type = mode >> 12
        assert mode_type in [0b1000, 0b1010, 0b1110]
        mode_perms = mode & 0b0000000111111111
        uid = int.from_bytes(content[idx+28: idx+32], "big")
        gid = int.from_bytes(content[idx+32: idx+36], "big")
        fsize = int.from_bytes(content[idx+36: idx+40], "big")
        sha = format(int.from_bytes(content[idx+40: idx+60], "big"), "040x")
        flags = int.from_bytes(content[idx+60: idx+62], "big")
        flag_assume_valid = (flags & 0b1000000000000000) != 0
        flag_extended = (flags & 0b0100000000000000) != 0
        assert not flag_extended
        flag_stage =  flags & 0b0011000000000000
        name_length = flags & 0b0000111111111111

        idx += 62

        if name_length < 0xFFF:
            assert content[idx + name_length] == 0x00
            raw_name = content[idx:idx+name_length]
            idx += name_length + 1
        else:
            print(f"Notice: Name is 0x{name_length:X} bytes long.")
            null_idx = content.find(b'\x00', idx + 0xFFF)
            raw_name = content[idx: null_idx]
            idx = null_idx + 1

        name = raw_name.decode("utf8")

        idx = 8 * ceil(idx / 8)

        entries.append(GITIndexEntry(ctime=(ctime_s, ctime_ns),
                                     mtime=(mtime_s,  mtime_ns),
                                     dev=dev,
                                     ino=ino,
                                     mode_type=mode_type,
                                     mode_perms=mode_perms,
                                     uid=uid,
                                     gid=gid,
                                     fsize=fsize,
                                     sha=sha,
                                     flag_assume_valid=flag_assume_valid,
                                     flag_stage=flag_stage,
                                     name=name))

    return GITIndex(version=version, entries=entries)



from helpers.repo.helpers import repo_file
from common.objects import GITObject
from common.commit.commit_obj import GITCommit
from common.tree.tree_obj import GITTree
from common.tree.tree_helper import GITTreeLeaf
from stage.indexfile import GITIndexEntry

import configparser
import os

def branch_get_active(repo):
    with open(repo_file(repo, "HEAD"), "r") as f:
        head = f.read()

    if head.startswith("ref: refs/heads/"):
        return(head[16:-1])
    else:
        return False
    
def commit_create(repo, tree, parent, author, timestamp, message):
    commit = GITCommit()

    commit.kvlm[b"tree"] = tree.encode("ascii")
    if parent:
        commit.kvlm[b"parent"] = parent.encode("ascii")

    message = message.strip() + "\n"
    offset = int(timestamp.astimezone().utcoffset().total_seconds())
    hours = offset // 3600
    minutes = (offset % 3600) // 60
    tz = "{}{:02}{:02}".format("+" if offset > 0 else "-", hours, minutes)

    author = author + timestamp.strftime(" %s ") + tz
    commit.kvlm[b"author"] = author.encode("utf8")
    commit.kvlm[b"committer"] = author.encode("utf8")
    commit.kvlm[None] = message.encode("utf8")

    return GITObject.object_write(commit, repo)

def tree_from_index(repo, index):
    contents = dict()
    contents[""] = list()

    for entry in index.entries:
        dirname = os.path.dirname(entry.name)
        key = dirname 
        while key != "":
            if not key in contents:
                contents[key] = list()
            key = os.path.dirname(key)
        contents[dirname].append(entry)      

    sorted_paths = sorted(contents.keys(), key=len, reverse=True)  
    sha = None

    for path in sorted_paths:
        tree = GITTree()
        for entry in contents[path]:
            if isinstance(entry, GITIndexEntry):
                leaf_mode = f"{entry.mode_type:02o}{entry.mode_perms:04o}".encode("ascii")
                leaf = GITTreeLeaf(mode = leaf_mode, path=os.path.basename(entry.name), sha=entry.sha)
            else: 
                leaf = GITTreeLeaf(mode = b"040000", path=entry[0], sha=entry[1])
            tree.items.append(leaf)
        sha = GITObject.object_write(tree, repo)

        parent = os.path.dirname(path)
        base = os.path.basename(path) 
        contents[parent].append((base, sha))
    
    return sha

def gitconfig_user_get(config):
    if "user" in config:
        if "name" in config["user"] and "email" in config["user"]:
            return f"{config['user']['name']} <{config['user']['email']}>"
    return None

def gitconfig_read():
    xdg_config_home = os.environ["XDG_CONFIG_HOME"] if "XDG_CONFIG_HOME" in os.environ else "~/.config"
    configfiles = [
        os.path.expanduser(os.path.join(xdg_config_home, "git/config")),
        os.path.expanduser("~/.gitconfig")
    ]

    config = configparser.ConfigParser()
    config.read(configfiles)
    return config


from helpers.repo.helpers import repo_file, repo_dir, repo_find
import os

def cmd_commits(args):
    repo = repo_find()
    my_dir = repo_dir(repo, "objects")
    for root, dirs, files in os.walk(my_dir):
        for filename in files:
            print(os.path.join(root, filename))



import os
from stage.readwrite import index_read, index_write

def rm(repo, paths, delete=True, skip_missing=False):
    index = index_read(repo)

    worktree = repo.worktree + os.sep

    abspaths = set()

    for path in paths:
        abspath = os.path.abspath(path)
        if abspath.startswith(worktree):
            abspaths.add(abspath)
        else:
            raise Exception(f"Cannot remove paths outside of worktree: {paths}")

    kept_entries = list()
    remove = list()

    for e in index.entries:
        full_path = os.path.join(repo.worktree, e.name)

        if full_path in abspaths:
            remove.append(full_path)
            abspaths.remove(full_path)
        else:
            kept_entries.append(e) 

    if len(abspaths) > 0 and not skip_missing:
        raise Exception(f"Cannot remove paths not in the index: {abspaths}")

    if delete:
        for path in remove:
            os.unlink(path)

    index.entries = kept_entries
    index_write(repo, index)




from common.objects import GITObject

def log_graphiz(repo, sha, seen):
    if sha in seen:
        return
    seen.add(sha)

    commitObj = GITObject.object_read(repo, sha)
    message = commitObj.kvlm[None].decode("utf8").strip()
    message = message.replace("\\", "\\\\")
    message = message.replace("\"", "\\\"")

    if "\n" in message: 
        message = message[:message.index("\n")]

    print(f"id: {sha[0:7]} | message: {message}")

    if not b'parent' in commitObj.kvlm:
        return

    parents = commitObj.kvlm[b'parent']

    if type(parents) != list:
        parents = [parents]

    for p in parents:
        p = p.decode("ascii")
        log_graphiz(repo, p, seen)



from common.objects import GITObject
from common.tree.tree_helper import tree_parse, tree_serialize

class GITTree(GITObject):
    obj_type = b'tree'

    def deserialize(self, data):
        self.items = tree_parse(data)
    
    def serialize(self):
        return tree_serialize(self)
    
    # old method
    # def __init__(self):
    #     self.items = list()
    
    def __init__(self, data=None):
        super().__init__(data)
        if not hasattr(self, 'items'):
            self.items = list()



from common.objects import GITObject
from common.commit.commit_helper import kvlm_parse, kvlm_serialize

class GITCommit(GITObject):
    obj_type = b'commit'

    def deserialize(self, data):
        self.kvlm = kvlm_parse(data)

    def serialize(self):
        return kvlm_serialize(self.kvlm)
    
    # old method
    # def __init__(self):
    #     self.kvlm = dict()

    def __init__(self, data=None):
        super().__init__(data)   # calls GITObject.__init__(data)
        if not hasattr(self, 'kvlm'):
            self.kvlm = dict()


import re
import os
from common.blob.blob_obj import GITBlob
from common.tag.tag_obj import GITTag
from common.tree.tree_obj import GITTree
from common.commit.commit_obj import GITCommit
from common.objects import GITObject
from helpers.repo.helpers import repo_dir
from helpers.refs.helpers import ref_resolve


def object_hash(fd, fmt, repo=None):
    data = fd.read()

    match fmt:
        case b'commit' : obj=GITCommit(data)
        case b'tree'   : obj=GITTree(data)
        case b'tag'    : obj=GITTag(data)
        case b'blob'   : obj=GITBlob(data)
        case _: raise Exception(f"Unknown type {fmt}!")

    return GITObject.object_write(obj, repo)

# helpers
def object_resolve(repo, name):
    candidates = list()
    hashRE = re.compile(r"^[0-9A-Fa-f]{4,40}$")

    if not name.strip():
        return None

    if name == "HEAD":
        return [ ref_resolve(repo, "HEAD") ]

    if hashRE.match(name):
        name = name.lower()
        prefix = name[0:2]
        path = repo_dir(repo, "objects", prefix, mkdir=False)
        if path:
            rem = name[2:]
            for f in os.listdir(path):
                if f.startswith(rem):
                    candidates.append(prefix + f)


    as_tag = ref_resolve(repo, "refs/tags/" + name)
    if as_tag: 
        candidates.append(as_tag)

    as_branch = ref_resolve(repo, "refs/heads/" + name)
    if as_branch:
        candidates.append(as_branch)

    as_remote_branch = ref_resolve(repo, "refs/remotes/" + name)
    if as_remote_branch:
        candidates.append(as_remote_branch)

    return candidates


def object_find(repo, name, obj_type=None, follow=True):
    sha = object_resolve(repo, name)

    if not sha:
        raise Exception(f"No such reference {name}.")

    if len(sha) > 1:
        raise Exception("That's weird")
    

    sha = sha[0]

    if not sha:
        return None

    if not obj_type:
        return sha
    
    while True:
        obj = GITObject.object_read(repo, sha)


        if obj.obj_type == obj_type:
            return sha
        
        if not follow:
            return None
        
        if obj.obj_type == b'tag':
            sha = obj.kvlm[b'object'].decode("ascii")
        elif obj.obj_type == b'commit' and obj_type == b'tree':
            sha = obj.kvlm[b'tree'].decode("ascii")
        else:
            return None

def tree_to_dict(repo, ref, prefix=""):
    ret = dict()
    tree_sha = object_find(repo, ref, obj_type=b"tree")

    if not tree_sha:
        return ret

    tree = GITObject.object_read(repo, tree_sha)

    for leaf in tree.items:
        full_path = os.path.join(prefix, leaf.path)
        is_subtree = leaf.mode.startswith(b'04')
        if is_subtree:
            ret.update(tree_to_dict(repo, leaf.sha, full_path))
        else:
            ret[full_path] = leaf.sha
    return ret



import os
from common.repo import GITRepository
import configparser
from termcolor import cprint

def repo_create(path):
    repo = GITRepository(path, True)

    if os.path.exists(repo.worktree):
        if not os.path.isdir(repo.worktree):
            cprint("The path that you provided is not a directory brother", "red")
            return None
        if os.path.exists(repo.gitdir):
            cprint("zyra repository is already initialized (.git folder is present)", "red")
            cprint("execute (rm -rf .git) to delete the folder", "yellow")
            return None
    else :
        os.makedirs(repo.worktree)
    
    repo_dir(repo, "branches", mkdir=True)
    repo_dir(repo, "objects", mkdir=True)
    repo_dir(repo, "refs", "tags", mkdir=True)
    repo_dir(repo, "refs", "heads", mkdir=True)

    with open(repo_file(repo, "description"), "w") as f:
        f.write("Unnamed repository yo; edit this file to name the repo")
    
    with open(repo_file(repo, "config"), "w") as f:
        ret = configparser.ConfigParser()
        ret.add_section("core")
        ret.set("core", "repositoryformatversion", "0")
        ret.set("core", "filemode", "false")
        ret.set("core", "bare", "false")
        ret.write(f)

    with open(repo_file(repo, "HEAD"), 'w') as f:
        f.write("ref: refs/heads/master\n")

    return repo


def repo_path(repo, *path):
    return os.path.join(repo.gitdir, *path)

def repo_file(repo, *path, mkdir=False):
    if repo_dir(repo, *path[:-1], mkdir=mkdir):
        return repo_path(repo, *path)

def repo_dir(repo, *path, mkdir=False):
    path = repo_path(repo, *path)

    if os.path.exists(path):
        if (os.path.isdir(path)):
            return path
        else:
            raise Exception(f"Not a directory {path}")

    if mkdir:
        os.makedirs(path)
        return path
    else:
        return None

def repo_find(path=".", required=True):
    path = os.path.realpath(path)
    if os.path.isdir(os.path.join(path, ".git")):
        return GITRepository(path)
    
    parent = os.path.realpath(os.path.join(path, ".."))
    if parent == path:
        if required:
            raise Exception("No git directory.")
        else:
            return None
    
    return repo_find(parent, required)



import os
from helpers.repo.helpers import repo_file
from helpers.objects.helpers import object_hash, object_find
from helpers.objects.helpers import tree_to_dict
from termcolor import cprint

def cmd_status_branch(repo):
    with open(repo_file(repo, "HEAD"), "r") as f:
        head = f.read()
    
    if head.startswith("ref: refs/heads/"):
        cprint(f"On branch {head[16:-1]}", "blue")
        print()
    else:
        print(f"Currently in detached mode : {object_find(repo, "HEAD")}")

def cmd_status_head_index(repo, index):
    # staging area is index
    cprint("Changes staged for commit:", "blue")
    head = tree_to_dict(repo, "HEAD")
    for entry in index.entries:
        if entry.name in head:
            if head[entry.name] != entry.sha:
                print("modified: ", entry.name)
            del head[entry.name]
        else:
            print("added: ", entry.name)

    for entry in head.keys():
        print("deleted: ", entry)

    print()

def cmd_status_index_worktree(repo, index):
    cprint("Changes not staged for commit:", "blue")

    gitdir_prefix = repo.gitdir + os.path.sep

    all_files = list()

    for (root, _, files) in os.walk(repo.worktree, True):
        if root==repo.gitdir or root.startswith(gitdir_prefix):
            continue
        for f in files:
            full_path = os.path.join(root, f)
            rel_path = os.path.relpath(full_path, repo.worktree)
            all_files.append(rel_path)

    for entry in index.entries:
        full_path = os.path.join(repo.worktree, entry.name)
        if not os.path.exists(full_path):
            print("  deleted: ", entry.name)
        else:
            stat = os.stat(full_path)

            ctime_ns = entry.ctime[0] * 10**9 + entry.ctime[1]
            mtime_ns = entry.mtime[0] * 10**9 + entry.mtime[1]
            if (stat.st_ctime_ns != ctime_ns) or (stat.st_mtime_ns != mtime_ns):
                with open(full_path, "rb") as fd:
                    new_sha = object_hash(fd, b"blob", None)
                    same = entry.sha == new_sha
                    if not same:
                        print(" modified:", entry.name)

        if entry.name in all_files:
            all_files.remove(entry.name)

    print()
    cprint("Untracked files:", "blue")

    for f in all_files:
        print(" ", f)



import argparse

argparser = argparse.ArgumentParser(description="The Git Learning Parser")
argsubparser = argparser.add_subparsers(dest="command", title="Commands")
argsubparser.required = True

# Initializes an empty git repository
init_parser = argsubparser.add_parser("init")
init_parser.add_argument("path")


# Prints out the content inside that hash file (sha file -> .git/objects/2a/as5272fs82dg)
# The internal function that calls this returns object that is type of the sha. (blob, commit etc). So when .serialize() it'll print.
catfile_parser = argsubparser.add_parser("cat-file")
catfile_parser.add_argument(
    "--obj-type", metavar="type", dest="type", choices=["blob", "commit", "tag", "tree"]
)
catfile_parser.add_argument("sha")


# Given a file it'll make it binary string in a proper format, compresses it stores in the sha path of that binary format (given -w option). It prints the sha hash object.
hashobject_parser = argsubparser.add_parser("hash-object")
hashobject_parser.add_argument(
    "-w",
    dest="write",
    action="store_true",
    help="Actually write the object into database",
)
hashobject_parser.add_argument("path")

# Takes the sha of a commit object and prints its contents
log_parser = argsubparser.add_parser("log", help="Display history of given commit")
log_parser.add_argument("commit", default="HEAD")


checkout_parser = argsubparser.add_parser("checkout")
checkout_parser.add_argument("commit", help="The commit or tree to checkout.")
checkout_parser.add_argument("path", help="The Empty directory to check on")

showref_parser = argsubparser.add_parser("show-ref", help="List references")

tag_parser = argsubparser.add_parser("tag", help="List and create tags")
tag_parser.add_argument(
    "-a",
    dest="create_tag_object",
    action="store_true",
    help="Whether to create a tag object",
)
tag_parser.add_argument("name", nargs="?", help="The new tag's name")
tag_parser.add_argument(
    "object", default="HEAD", nargs="?", help="The object the new tag will point to"
)

rev_parser = argsubparser.add_parser("rev-parse", help="Converts name to hash")
rev_parser.add_argument(
    "--wyag-type", dest="type", choices=["blob", "commit", "tag", "tree"]
)
rev_parser.add_argument("name")


# git status command
# On branch master

# Changes to be committed: (comparision of index file with HEAD (previous commit and staging area))

# Changes not staged for commit: (comparisin with staging area (index file) and working tree(workdir))

# Untracked files: (comparision of index file with HEAD (previous commit and staging area))
status_parser = argsubparser.add_parser("status", help="Print the status of the game")


rm_parser = argsubparser.add_parser("rm", help="Remove paths from the staging area")
rm_parser.add_argument("-n", dest="nd", action="store_true", help="unstage")
rm_parser.add_argument("path", nargs="+", help="Paths to remove")


add_parser = argsubparser.add_parser("add", help="Adding files to staging area")
add_parser.add_argument("path", nargs="+", help="Paths going to staging area")


commit_parser = argsubparser.add_parser("commit", help="Finally commit it goddammit")
commit_parser.add_argument(
    "-m",
    metavar="message",
    dest="message",
    help="Message to associate with this commit",
)

commits_parser = argsubparser.add_parser("all-commits", help="Display all the commits")

branch_parser = argsubparser.add_parser("branch", help="Look at your current branch")



class GITIndex(object):
    entries = []
    version = None

    def __init__(self, version=2, entries=None):
        if not entries:
            entries = list()
        self.version = version
        self.entries = entries

# Index file : header(DIRC + format version number + number of entries) and entries
class GITIndexEntry(object):
    def __init__(self, ctime=None, mtime=None, dev=None, ino=None,
                 mode_type=None, mode_perms=None, uid=None, gid=None,
                 fsize=None, sha=None, flag_assume_valid=None,
                 flag_stage=None, name=None):
        
        self.ctime = ctime
        self.mtime = mtime
        self.dev = dev
        self.ino = ino
        self.mode_type = mode_type
        self.mode_perms = mode_perms
        self.uid = uid
        self.gid = gid
        self.fsize = fsize
        self.sha = sha
        self.flag_assume_valid = flag_assume_valid
        self.flag_stage = flag_stage
        self.name = name



from common.objects import GITObject
import os

def tree_checkout(repo, tree, path):
    for item in tree.items:
        obj = GITObject.object_read(repo, item.sha)
        dest = os.path.join(path, item.path)

        if obj.obj_type == b'tree':
            os.makedirs(dest)
            tree_checkout(repo, obj, dest)
        elif obj.obj_type == b'blob':
            with open(dest, "wb") as f:
                f.write(obj.blobdata)



from cmds.rm import rm
import os
from stage.indexfile import GITIndexEntry
from stage.readwrite import index_read, index_write
from helpers.objects.helpers import object_hash

def add(repo, paths, delete=True, skip_missing=False):
    if paths[0] == ".":
        all_files = list()
        gitdir_prefix = repo.gitdir + os.path.sep
        for (root, _, files) in os.walk(repo.worktree, True):
            if root==repo.gitdir or root.startswith(gitdir_prefix):
                continue
            for f in files:
                full_path = os.path.join(root, f)
                rel_path = os.path.relpath(full_path, repo.worktree)
                all_files.append(rel_path)
                
        paths = all_files

    rm(repo, paths, delete=False, skip_missing=True)
    worktree = repo.worktree + os.sep

    cleanpaths = set()
    for path in paths:
        p = os.path.abspath(path)
        relpath = os.path.relpath(p, repo.worktree)
        if not (p.startswith(worktree) and os.path.isfile(p)):
            raise Exception(f"Not a file, or outside the worktree: {paths}")

        cleanpaths.add((p, relpath))

    index = index_read(repo)



    for (abspath, relpath) in cleanpaths:
        with open(abspath, "rb") as f:
            sha = object_hash(f, b"blob", repo)
            stat = os.stat(abspath)
            ctime_s = int(stat.st_ctime)
            ctime_ns = stat.st_ctime_ns % 10**9
            mtime_s = int(stat.st_mtime)
            mtime_ns = stat.st_mtime_ns % 10**9

            entry = GITIndexEntry(ctime=(ctime_s, ctime_ns), mtime=(mtime_s, mtime_ns), dev=stat.st_dev, ino=stat.st_ino,
                                  mode_type=0b1000, mode_perms=0o644, uid=stat.st_uid, gid=stat.st_gid,
                                  fsize=stat.st_size, sha=sha, flag_assume_valid=False,
                                  flag_stage=False, name=relpath)
            index.entries.append(entry)

    index_write(repo, index)



from common.objects import GITObject

class GITBlob(GITObject):
    obj_type = b'blob'

    def serialize(self):
        return self.blobdata
    
    def deserialize(self, data):
        self.blobdata = data




from setuptools import setup, find_packages


setup(
    name="zyra-tool",
    version="1.0.8",
    author="Karthik Rishinarada",
    author_email="karthikdmy11135@gmail.com",
    description="zyra: A version control system that is built from scratch in python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/karthik11135/zyra",
    packages=find_packages(),
    py_modules=["zyra", "argparsing", "main"],
    include_package_data=True,
    install_requires=[
        "termcolor>=2.2.0",
    ],
    entry_points={
        "console_scripts": [
            "zyra=zyra:main",
        ],
    },
    python_requires=">=3.10",
)



from helpers.repo.helpers import repo_find, repo_create
from cmds.add import add
from common.blob.blob_obj import GITBlob
from common.tag.tag_obj import GITTag
from common.tree.tree_obj import GITTree
from common.commit.commit_obj import GITCommit
from common.objects import GITObject
from helpers.objects.helpers import *
from cmds.checkout import tree_checkout
from helpers.refs.helpers import *
from cmds.log import log_graphiz
from cmds.tag import tag_create
from cmds.add import add
from cmds.rm import rm
from cmds.commit import *
from cmds.checkout import *
from stage.readwrite import index_read
from cmds.status import *
from termcolor import cprint

import sys
from datetime import datetime



def cmd_branch(args):
    repo = repo_find()
    head_file = repo_file(repo, "HEAD")
    print("branches:")
    with open(head_file, "r") as f:
        data = f.read()

    if data.startswith("ref:"):
        my_branch = data[16:-1]
        cprint(f" * {my_branch.strip()}", "green")
    else:
        my_branch = data[:-1]
        cprint(f" detached state: * {my_branch.strip()}", "green")

    branches_path = repo_dir(repo, "refs", "heads")
    for bs in os.listdir(branches_path):

        if bs == my_branch.strip():
            continue
        print("   " + bs)


def cmd_add(args):
    repo = repo_find()
    add(repo, args.path)


def cmd_checkout(args):
    repo = repo_find()
    sha = args.commit
    path = args.path

    obj = GITObject.object_read(repo, object_find(repo, sha))

    if obj.obj_type == b"commit":
        obj = GITObject.object_read(repo, obj.kvlm[b"tree"].decode("ascii"))

    if os.path.exists(path):
        if not os.path.isdir(path):
            raise Exception("The provided path is not a directory")
        if os.listdir(path):
            pass
    else:
        os.makedirs(path)

    tree_checkout(repo, obj, os.path.realpath(path))


def cmd_init(args):
    if repo_create(args.path):
        cprint("zyra repository is succesfully initialised", "light_blue")


def cmd_cat_file(args):
    repo = repo_find()
    t = args.type
    obj_type = t.encode() if t else None
    obj = GITObject.object_read(repo, object_find(repo, args.sha, obj_type=obj_type))
    print(obj.obj_type)
    sys.stdout.buffer.write(obj.serialize())


def cmd_hash_obj(args):
    write = args.write
    path = args.path
    if write:
        repo = repo_find()
    else:
        repo = None

    with open(path, "rb") as f:
        data = f.read()
        obj = GITBlob(data)
        sha = GITObject.object_write(obj, repo)
        print(sha)


def cmd_log(args):
    repo = repo_find()

    print("Logs logs logs")
    try:
        log_graphiz(repo, object_find(repo, args.commit), set())
        print("Logs ended here")
    except:
        cprint("You are supposed to provide the sha of a commit object only", "red")


def cmd_show_ref(args):
    repo = repo_find()
    ref_dict = ref_list(repo)
    print(ref_dict)
    show_ref(repo, ref_dict, prefix="refs")


def cmd_tag(args):
    repo = repo_find()

    if args.name:
        tag_create(
            repo, args.name, args.object, create_tag_object=args.create_tag_object
        )
    else:
        ref_dict = ref_list(repo)
        show_ref(repo, ref_dict, prefix="refs")


def cmd_rev_parse(args):
    if args.type:
        obj_type = args.type.encode()
    else:
        obj_type = None

    repo = repo_find()

    print(object_find(repo, args.name, obj_type=obj_type, follow=True))


def cmd_rm(args):
    repo = repo_find()
    nd = args.nd
    if nd:
        delete = False
    else:
        delete = True
    rm(repo, args.path, delete=delete)


def cmd_commit(args):
    repo = repo_find()
    index = index_read(repo)
    tree = tree_from_index(repo, index)
    commit = commit_create(
        repo,
        tree,
        object_find(repo, "HEAD"),
        gitconfig_user_get(gitconfig_read()),
        datetime.now(),
        args.message,
    )

    active_branch = branch_get_active(repo)
    if active_branch:
        with open(
            repo_file(repo, os.path.join("refs/heads/"), active_branch), "w"
        ) as f:
            f.write(commit + "\n")
    else:
        with open(repo_file(repo, "HEAD"), "w") as f:
            f.write("\n")


def cmd_status(args):
    repo = repo_find()
    index = index_read(repo)

    cmd_status_branch(repo)
    cmd_status_head_index(repo, index)
    cmd_status_index_worktree(repo, index)


def cmd_commits(args):
    repo = repo_find()
    my_dir = repo_dir(repo, "objects")
    my_shas = list()
    obj_dirs = list()
    i = 0
    for root, dirs, files in os.walk(my_dir):
        if dirs:
            obj_dirs = dirs
        if files:
            my_shas.append(obj_dirs[i] + files[0])
            i += 1

    for p in my_shas:
        obj = GITObject.object_read(repo, p)
        if obj.obj_type == b"commit":
            print(f"{p}: {obj.kvlm[None][:-1].decode()}")



# converts the commit raw string (ig its binary only) into dictionary
def kvlm_parse(raw, start=0, dct=None):
    if not dct:
        dct = dict()

    spc = raw.find(b' ', start)
    nl = raw.find(b'\n', start)

    if (spc < 0) or (nl < spc):
        assert nl == start
        dct[None] = raw[start+1:]
        return dct

    key = raw[start:spc]

    end = start
    while True:
        end = raw.find(b'\n', end+1)
        if raw[end+1] != ord(' '): break

    value = raw[spc+1:end].replace(b'\n ', b'\n')

    if key in dct:
        if type(dct[key]) == list:
            dct[key].append(value)
        else:
            dct[key] = [dct[key], value]
    else:
        dct[key]=value

    return kvlm_parse(raw, start=end+1, dct=dct)


# converts the commit dictionary into raw string
def kvlm_serialize(kvlm):
    ret = b''

    # Output fields
    for k in kvlm.keys():
        # Skip the message itself
        if k == None: continue
        val = kvlm[k]

        # Normalize to a list
        if type(val) != list:
            val = [ val ]

        print(k)
        # print(k_encode)

        for v in val:
            ret += k + b' ' + (v.replace(b'\n', b'\n ')) + b'\n'

    # Append message
    ret += b'\n' + kvlm[None]

    return ret



import os
from helpers.repo.helpers import repo_file, repo_dir

def ref_resolve(repo, ref):
    path = repo_file(repo, ref)
    if not os.path.isfile(path):
        return None
    
    with open(path, "r") as f:
        data = f.read()[:-1]

    if data.startswith("ref: "):
        return ref_resolve(repo, data[5:])
    else:
        return data
    
def ref_list(repo, path=None):
    if not path:
        path = repo_dir(repo, "refs")

    ret = dict()

    for f in sorted(os.listdir(path)):
        can = os.path.join(path, f)

        if os.path.isdir(can):
            ret[f] = ref_list(repo, can)
        else :
            ret[f] = ref_resolve(repo, can)

    return ret

def show_ref(repo, refs, with_hash=True, prefix=""):
    if prefix:
        prefix = prefix + '/'
    for k, v in refs.items():
        if type(v) == str and with_hash:
            print (f"{v} {prefix}{k}")
        elif type(v) == str:
            print (f"{prefix}{k}")
        else:
            show_ref(repo, v, with_hash=with_hash, prefix=f"{prefix}{k}")

def ref_create(repo, path, sha):
    with open(repo_file(repo, "refs", path), 'w') as f:
        f.write(sha + "\n")


class GITTreeLeaf(object):
    def __init__(self, mode, path, sha):
        self.mode = mode
        self.path = path
        self.sha = sha

def tree_parse(raw):
    pos = 0
    total = len(raw)
    ret = []
    while pos < total:
        pos, data = tree_parse_one(raw, pos)
        ret.append(data)
    return ret

def tree_leaf_sort_key(leaf):
    if leaf.mode.startswith(b"10"):
        return leaf.path
    else:
        return leaf.path + "/"

def tree_serialize(obj):
    obj.items.sort(key=tree_leaf_sort_key)

    ret = b""

    for o in obj.items:
        ret += o.mode
        ret += b" "
        ret += o.path.encode("utf8")
        ret += b"\x00"
        sha = int(o.sha, 16)
        ret += sha.to_bytes(20, byteorder="big")
    
    return ret

def tree_parse_one(raw, start = 0):
    x = raw.find(b' ', start)
    assert x-start == 5 or x-start == 6
    mode = raw[start:x]

    if len(mode) == 5:
        mode += b"0"

    y = raw.find(b'\x00', x)
    path = raw[x+1: y]

    raw_sha = int.from_bytes(raw[y+1: y+21], "big")
    sha = format(raw_sha, "040x")
    return y + 21, GITTreeLeaf(mode, path.decode("utf8"), sha)



import hashlib
import os
import zlib
from helpers.repo.helpers import repo_file

# An object is stored in this manner: 
# b'{typeof object}{size} \x00{content}' (obj type: {blob, commit, tag, tree})
class GITObject(object):
    def __init__(self, data=None):
        if data != None:
            self.deserialize(data)
        else:
            self.init()

    def serialize(self, repo):
        pass

    def deserialize(self, repo):
        pass

    def init(self):
        pass

    #   Takes the type object(GitBlob, GitCommit etc) and optional repo (whether to write the contents of the file in the sha path or not) and returns the sha of the content. 
    @staticmethod
    def object_write(obj, repo=None):
        # Forming metadata -> head space size content
        data = obj.serialize() # this gives the content of the object in the format of a b string
        head = obj.obj_type
        str_len = str(len(data)).encode()
        space = b' '
        zero_bin = b'\x00'

        final_str = head + space + str_len  + zero_bin + data
        
        sha = hashlib.sha1(final_str).hexdigest()

        if not repo:
            return sha

        path = repo_file(repo, "objects", sha[0:2], sha[2:], mkdir=True)

        print(path)

        if not os.path.exists(path):
            with open(path, "wb") as f:
                f.write(zlib.compress(final_str))
        
        return sha


    # takes the sha and returns the class according to the type of object. 
    @staticmethod
    def object_read(repo, sha):
        path = repo_file(repo, "objects", sha[0:2], sha[2:])

        if not os.path.isfile(path):
            raise Exception("No object exists here")
        
        with open(path, "rb") as f:
            decomp_str = zlib.decompress(f.read())
            # b'blob{size} \x00{content}' -> format of storage. 
            x = decomp_str.find(b' ')

            obj_type = decomp_str[0:x]
            y = decomp_str.find(b'\x00', x)

            content = decomp_str[y+1:]

            from common.blob.blob_obj import GITBlob
            from common.tag.tag_obj import GITTag
            from common.tree.tree_obj import GITTree
            from common.commit.commit_obj import GITCommit
            match obj_type:
                case b'commit': c=GITCommit
                case b'tree'  : c=GITTree
                case b'tag'   : c=GITTag
                case b'blob'  : c=GITBlob
                case _:
                    raise Exception("I dont know what object this is")
            
            return c(decomp_str[y+1:])
        

